from pyblast.constants import Constants as C
from uuid import uuid4
from pyblast.exceptions import SeqRecordValidationError
import networkx as nx
from pyblast.utils import is_circular
from collections.abc import Sized


class GraphDB(Sized):
    def __init__(self):
        self.graph = nx.DiGraph()
        self.mapping = {}

    @staticmethod
    def new_key():
        """Generate a new unique key"""
        return str(uuid4())

    def key(self, obj):
        """Model id to key. To ensure uniqueness of the models being added."""
        k = self.mapping.get(id(obj), None)
        return k

    def get_many(self, keys, default=C.NULL):
        """Get many objects from list of keys."""
        records = []
        for k in keys:
            records.append(self.get(k, default=default))
        return records

    def get(self, key, default=C.NULL):
        """Get object from a key"""
        if default is not C.NULL:
            if key not in self.graph:
                return default
            return self.graph.node[key]["object"]
        return self.graph.node[key]["object"]

    def add(self, obj):
        """Add an object"""
        if self.key(obj):
            return self.key(obj)
        else:
            key = self.new_key()
            self.graph.add_node(key, object=obj)
            self.mapping[id(obj)] = key
            return key

    def add_many(self, objects):
        """Add many objects"""
        keys = []
        for o in objects:
            keys.append(self.add(o))
        return keys

    def add_edge(self, k1, k2, **kwargs):
        """Add an edge"""
        return self.graph.add_edge(k1, k2, **kwargs)

    def to_dict(self):
        """Return key to object dictionary"""
        d = {}
        for key in self.graph:
            d[key] = self.get(key)
        return d

    def __len__(self):
        return len(self.to_dict())


class SeqRecordDB(GraphDB):
    def __init__(self):
        super().__init__()
        self._transformations = {}

    @staticmethod
    def is_circular(records):
        return is_circular(records)

    @property
    def records(self):
        """Return all records"""
        return self.to_dict()

    @classmethod
    def validate_records(cls, records):
        for r in records:
            cls.validate_circular(r)

    @staticmethod
    def validate_circular(r):
        """Ensure the record has a 'topology' of the expected type ['circular', 'linear']"""

        valid_keys = [C.CIRCULAR, C.LINEAR, C.TOPOLOGY]
        if all([k not in r.annotations for k in valid_keys]):
            raise SeqRecordValidationError(
                "SeqRecord {} is missing a any of the valid keys in its annotation: {keys}. This must be provided.\n"
                "`Use pyblast.utils.make_circular` or `pyblast.utils.make_linear` to annotate SeqRecords".format(
                    r, keys=valid_keys
                )
            )
        if C.LINEAR in r.annotations and C.CIRCULAR in r.annotations:
            if r.annotations[C.LINEAR] is r.annotations[C.CIRCULAR]:
                raise SeqRecordValidationError(
                    "SeqRecord {} has conflicting topology definitions for '{linear}' and '{circular}'".format(
                        r, circular=C.CIRCULAR, linear=C.LINEAR
                    )
                )

    def transform(self, parent_key, transform, transform_label):
        """
        Transform a record.

        :param parent_key:
        :type parent_key:
        :param transform:
        :type transform:
        :param transform_label:
        :type transform_label:
        :return:
        :rtype:
        """
        record = self.get(parent_key)
        new_record = transform(record)
        new_key = self.add(new_record)
        self.add_edge(
            parent_key,
            new_key,
            **{C.TRANSFORMATION: transform_label, C.PARENT: parent_key}
        )
        new_record._transform = transform_label
        new_record.id = new_key
        return new_key

    def add_with_transformation(self, record, transform, transform_label):
        """
        Add a record and its transformation to the database. If the original
        record exists in the database, then that record is used. If the original
        record has had the same type of transformation applied, no new records are
        added and the previously transformed key is returned. Else, the `transform` function
        is applied and the new record is added, with a trace back to the original record.

        :param record: the record to apply a transformation
        :type record: SeqRecord
        :param transform: the transformation function that takes in a single argument, the record
        :type transform: callable
        :param transform_label: the transformation label of the transform function
        :type transform_label: str
        :return: the new key that refers to the transformed record
        :rtype: str
        """
        key = self.add(record)
        transform_key = (key, transform_label)
        if transform_key in self._transformations:
            return self._transformations[transform_key]
        else:
            new_key = self.transform(key, transform, transform_label)
            self._transformations[transform_key] = new_key
            return new_key

    def add_many_with_transformations(self, records, transform, transform_label):
        """
        Transform many records. See `add_with_transformation` for usage.
        """
        return [
            self.add_with_transformation(r, transform, transform_label) for r in records
        ]

    def get_origin_key(self, key):
        """
        Trace the record at 'key' back to its origin through
         its transformation trace and return the origin's key.

        :param key: the key of the record
        :type key: str
        :return: the original record key
        :rtype: str
        """
        tree = nx.bfs_tree(self.graph, key, reverse=True)
        roots = [n for n, d in dict(tree.out_degree()).items() if d == 0]
        return roots[0]

    def get_origin(self, key):
        """
        Trace the record at 'key' back to its origin through
         its transformation trace and return the origin record.

        :param key: the key of the record
        :type key: str
        :return: the original record
        :rtype: SeqRecord
        """
        return self.get(self.get_origin_key(key))

    def add(self, record, validate=True):
        """Add a record"""
        if self.key(record):
            return self.key(record)
        else:
            if validate:
                self.validate_records([record])
            return super().add(record)

    def add_many(self, records):
        """Add many records"""
        self.validate_records(records)
        return super().add_many(records)



"""cmd.py"""

import shlex
from subprocess import check_output, STDOUT, CalledProcessError


def run_cmd_str(cmd_str):
    """Runs a command from a string"""
    args = shlex.split(cmd_str)
    try:
        output = check_output(args, stderr=STDOUT)
    except CalledProcessError as error:
        error_message = ">>> Error while executing:\n{}\n>>> Returned with error:\n{}".format(
            cmd_str, str(error.output)
        )
        error_message += (
            "\nParameter info: https://www.ncbi.nlm.nih.gov/books/NBK279684/"
        )
        raise Exception(error_message)

    # output = subprocess.Popen(args, shell=False)
    # output.wait()
    return output.decode()


def run_cmd(cmd, **kwargs):
    """Run a command using parameters kwargs"""
    run_cmd_str(dict_to_cmd(cmd, **kwargs))


def dict_to_cmd(cmd, **kwargs):
    """Create a command string for cmd and parameters 'kwargs'"""
    return cmd + " " + " ".join(["-{} {}".format(k, kwargs[k]) for k in kwargs])



import fire
from pyblast.blast_bin import BlastWrapper


def status(path=None):
    """
    Checks status of blast.

    :param path: installation path
    :type path: basestring
    """
    b = BlastWrapper(path)
    path = b.is_installed()
    if not path:
        print("FALSE: Blast is not installed.")
    else:
        print("TRUE: Blast is installed at {}".format(path))


def install(email=None, platform=None, path=None):
    """
    Installs blast

    :param email: users email
    :type email: basestring
    :param platform: platform (mac, linux, etc.)
    :type platform: basestring
    :param path: installation path
    :type path: basestring
    :return: None
    :rtype: None
    """
    b = BlastWrapper(path)
    if email is None and platform is None:
        b.ask_to_install()
    else:
        return b.install(email, platform)


def main():
    return fire.Fire({"install": install, "status": status})


def entrypoint():
    return main()


if __name__ == "__main__":
    main()



"""

.. module:: pyblast

Submodules
==========

.. autosummary::
    :toctree: _autosummary

    BioBlast
    JSONBlast
    BioBlastFactory
    utils.Span
    utils
"""

from .blast import BioBlast, JSONBlast, BioBlastFactory
from .__version__ import __version__, __title__, __authors__, __homepage__, __repo__
from .cli import entrypoint



class PyBlastWarning(Warning):
    """A generic warning for pyblastbio"""


class PyBlastException(Exception):
    """A generic exception for pyblastbio"""


class SeqRecordValidationError(Exception):
    """An error if a Bio.SeqRecord instance is invalid."""



# __version__.py
# autogenerated by keats 0.2.26
__version__ = "0.4.1"
__name__ = "pyblastbio"
__title__ = "pyblastbio"
__authors__ = ["Justin Vrana <justin.vrana@gmail.com>"]
__repo__ = None
__homepage__ = "https://www.github.com/jvrana/pyblast"
__description__ = ""



from Bio.SeqFeature import SeqFeature
from Bio.SeqRecord import SeqRecord
from uuid import uuid4
from Bio.Seq import Seq
from pyblast.utils import new_feature_location
from pyblast.exceptions import PyBlastException


class JSONParser(object):
    @staticmethod
    def clean_data(data):
        return {k: v for k, v in data.items() if v is not None}

    @classmethod
    def JSON_to_SeqFeature(cls, data, length=None):
        """Convert JSON to a Bio.SeqFeature object"""
        start = data["start"]
        end = data["end"]
        strand = data["strand"]
        return SeqFeature(
            location=new_feature_location(start, end, strand=strand, length=length),
            type=data["type"],
            id=data["name"],
        )

    @classmethod
    def __JSON_to_SeqRecord(cls, data):
        if not issubclass(type(data), dict) and not issubclass(type(data), list):
            raise PyBlastException(
                "Data must be a json object but found a '{}'".format(type(data))
            )
        data = cls.clean_data(data)
        return SeqRecord(
            seq=Seq(data["bases"]),
            annotations={"circular": data.get("circular", False)},
            features=[
                cls.JSON_to_SeqFeature(f, len(data["bases"]))
                for f in data.get("features", [])
            ],
            description=data.get("description", ""),
            name=data["name"],
            id=data.get("id", str(uuid4())),
        )

    @classmethod
    def JSON_to_SeqRecords(cls, datalist):
        """Convert JSON to SeqRecords"""
        records = []
        if isinstance(datalist, list):
            for data in datalist:
                records.append(cls.__JSON_to_SeqRecord(data))
        else:
            records.append(cls.__JSON_to_SeqRecord(datalist))
        return records



"""bioblastfactory"""


from pyblast.blast import BioBlast
from pyblast.blast.seqdb import SeqRecordDB
from pyblast.utils import clean_records
from typing import List, Tuple
from Bio.SeqRecord import SeqRecord
from pyblast.log import logger


class BioBlastFactory(object):
    """
    Instance factory for :class:`BioBlast` instance that share the same :class:`SeqRecordDB`.

    Usage:

    ::

        factory = BioBlastFactory()
        factory["plasmid_templates"] = records1
        factory["primers"] = records2
        factory["queries"] = records3

        blast1 = factory("plasmid_templates", "queries")
        blast2 = factory("primers", "queries")
    """

    def __init__(self, seq_db=None, span_origin=True, config=None):
        """
        Initialize a new BioBlastFactory.

        :param seq_db: the optional SeqRecordDB. If not provided, a new one will be created.
        :type seq_db: SeqRecordDB
        """
        if seq_db is None:
            self.db = SeqRecordDB()
        else:
            self.db = seq_db
        self.span_origin = span_origin
        self.record_groups = {}
        self.logger = logger(self)
        self.config = config

    def __setitem__(self, record_group_name: str, records: List[SeqRecord]):
        """
        See add_records.
        """
        self.add_records(records, record_group_name)

    def __getitem__(self, record_group_name):
        return self.record_groups[record_group_name]

    def add_records(
        self, records: List[SeqRecord], record_group_name: str
    ) -> Tuple[List[str], List[SeqRecord]]:
        """
        Add records to the SeqRecordDB by keyword.

        :param records:
        :type records:
        :param record_group_name:
        :type record_group_name:
        :return:
        :rtype:
        """
        self.logger.info(
            "Adding {} records to '{}'".format(len(records), record_group_name)
        )
        clean_records(records)
        keys, records = BioBlast.add_records(
            records, self.db, span_origin=self.span_origin
        )
        if record_group_name:
            self.record_groups.setdefault(record_group_name, list())
            self.record_groups[record_group_name] += records
        return keys, records

    def __call__(self, subject_key, query_key, config=None):
        """
        Create a new BioBlast instance with the factory's SeqRecordDB.
        
        :param subject_key: the subject key
        :type subject_key: str
        :param query_key: the query key
        :type query_key: str
        :param config: BioBlast config
        :type config: dict
        :return:
        :rtype:
        """
        self.logger.info("new Blast({}, {})".format(subject_key, query_key))
        if isinstance(subject_key, str):
            subjects = self.record_groups[subject_key]
        else:
            subjects = []
            for key in subject_key:
                if issubclass(type(key), SeqRecord):
                    subjects += key
                else:
                    subjects += self.record_groups[key]
        if isinstance(query_key, str):
            queries = self.record_groups[query_key]
        else:
            queries = []
            for key in query_key:
                if issubclass(type(key), SeqRecord):
                    queries += key
                else:
                    queries += self.record_groups[key]
        if config is None:
            config = {}
        if self.config:
            config.update(self.config)
        return BioBlast(
            subjects=subjects,
            queries=queries,
            seq_db=self.db,
            span_origin=self.span_origin,
            config=config,
        )



"""

.. module:: blast

Submodules
==========

.. autosummary::
    :toctree: _autosummary

    bioblastfactory
"""
from pyblast.blast.blast import BioBlast, JSONBlast, TmpBlast, BlastBase
from .bioblastfactory import BioBlastFactory



"""
blast.py

More information: https://www.ncbi.nlm.nih.gov/guide/howto/run-blast-local/
BLAST User Manual: https://www.ncbi.nlm.nih.gov/books/NBK1762/
"""

import json
import os
import shutil
import tempfile
import typing
from uuid import uuid4

from Bio import SeqIO
from Bio.SeqRecord import SeqRecord
from more_itertools import unique_everseen

from pyblast.blast_bin import BlastWrapper
from pyblast.constants import Constants as C
from pyblast.exceptions import PyBlastException, PyBlastWarning
from pyblast.utils import Span
from pyblast.utils import (
    glob_fasta_to_tmpfile,
    records_to_tmpfile,
    clean_records,
    force_unique_record_ids,
)
from pyblast.utils import run_cmd
from .blast_parser import BlastResultParser
from .json_parser import JSONParser
from .seqdb import SeqRecordDB


class BlastBase(object):
    """
    A Blast initializer for running blast searches against subjects contained in a directory. Sequences
    are kept on the local machine.

    Note: This is not intended to be the endpoint. You should only use temporary files to run the blast search.
    """

    outfmt = [
        "7",
        "qacc",
        "sacc",
        "score",
        "evalue",
        "bitscore",
        "length",
        "nident",
        "gapopen",
        "gaps",
        "qlen",
        "qstart",
        "qend",
        "slen",
        "sstart",
        "send",
        "sstrand",
        "qseq",
        "sseq",
    ]

    blast_config = {"outfmt": '"{0}"'.format(" ".join(outfmt))}

    def __init__(
        self,
        db_name,
        subject_path,
        query_path,
        db_output_directory,
        results_out_path,
        output_formatter=None,
        config=None,
    ):
        """
        A Blast initializer for running blast searches.

        :param db_name: Name for database file structure. This name will be appended to all db
        files that blast creates.
        :param subject_path: single fsa file containing the subject sequences
        :param query_path: Location of the fasta or genbank file containing the query
        :param db_output_directory: Location to store database related files
        :param results_out_path: Path to store the results.out file. Path can be absolute or relative.
        :param config: Additional configurations to run for the blast search (see
        https://www.ncbi.nlm.nih.gov/books/NBK279682/)
        """

        # name of the database
        self.db_name = None
        self.db_output_directory = None
        self.results_out_path = None
        self.set_paths(db_name, db_output_directory, results_out_path)

        # add executable to the system path
        blast_wrapper = BlastWrapper()
        if not blast_wrapper.is_installed():
            blast_wrapper.ask_to_install()

        self.query_path = query_path
        self.subject_path = subject_path

        # setup config
        if config is None:
            config = {}
        self._config = dict(config)

        # build the configuration
        if output_formatter is None:
            self.outfmt = BlastBase.outfmt
        else:
            self.outfmt = output_formatter

        # list of expected fields in the result
        self.fields = ()

        # the raw results from running blast from the command line
        self.raw_results = None

        # the parsed results as a dictionary
        self.results = None

    def set_paths(self, db_name, db_out_dir, results_out_path):
        self.db_name = db_name
        if db_out_dir:
            self.db_output_directory = os.path.abspath(db_out_dir)
        else:
            self.db_output_directory = db_out_dir
        if results_out_path:
            self.results_out_path = os.path.abspath(results_out_path)
        else:
            self.results_out_path = results_out_path

    @property
    def db(self):
        return os.path.join(self.db_output_directory, self.db_name)

    @property
    def config(self):
        out_config = {"outfmt": '"{0}"'.format(" ".join(self.outfmt))}
        out_config.update(self._config)
        return out_config

    def update_config(self, config):
        self._config.update(config)

    def validate_files(self):
        """
        Validate the directories and query files

        :return: None
        :rtype: None
        :raises: PyBlastException if filepaths are invalid or if query sequence has more than one sequence
        """

        def _is_file(myfile):
            return os.path.isfile(os.path.abspath(myfile))

        def _is_dir(mydir):
            return os.path.isdir(os.path.abspath(mydir))

        outdir = os.path.dirname(self.results_out_path)
        errors = []
        for file_ in [self.query_path, self.subject_path]:
            if not _is_file(file_):
                errors.append("File not found: {}".format(file_))
        for dir_ in [outdir, self.db_output_directory]:
            if not _is_dir(dir_):
                errors.append("Directory not found {}".format(dir_))
        if len(errors) > 0:
            raise PyBlastException("\n".join(errors))

        # seq = PySequence.open(self.query_path)
        # if len(seq) == 0:
        #     raise PyBlastException("Query path \"{}\" has no sequences".format(self.query_path))
        # elif len(seq) > 1:
        #     raise PyBlastException("Query path \"{}\" has more than one sequence.".format(self.query_path))

    def create_config(self):
        """Create a configuration dictionary"""
        config_dict = {
            "db": self.db,
            "out": self.results_out_path,
            "query": self.query_path,
        }
        config_dict.update(self.config)
        return config_dict

    def blastn(self):
        """Alias of 'quick_blastn'"""
        self.quick_blastn()

    def quick_blastn(self):
        """Produce database, run blastn protocol, & parse results """
        self.makedb()
        self._run_blastn()
        self.closedb()
        return self.parse_results()

    def blastn_short(self):
        """Alias of 'quick_blastn_short'"""
        self.quick_blastn_short()

    def quick_blastn_short(self):
        """Produce database, run blastn short protocol, & parse results """
        self.makedb()
        self._run_blastn_short()
        self.closedb()
        return self.parse_results()

    def _run_blastn(self):
        return self._run("blastn")

    def _run_blastn_short(self):
        return self._run("blastn", task="blastn-short")

    def _run(self, cmd, **config_opts):
        """Run the blastn using the current configuration"""
        config = self.create_config()
        config.update(config_opts)
        self.run_cmd(cmd, **config)
        with open(self.results_out_path, "rU") as handle:
            self.raw_results = handle.read()
        return self.raw_results

    # Wrapper for the util.run_cmd
    @staticmethod
    def run_cmd(cmd, **kwargs):
        """Wrapper for utils.run_cmd"""
        run_cmd(cmd, **kwargs)

    @staticmethod
    def _fasta_details(path):
        seqs = list(SeqIO.parse(path, format="fasta"))
        tot_bps = sum([len(s) for s in seqs])
        return {"num_sequence": len(seqs), "total_bps": tot_bps}

    def makedb(self, verbose=False):
        """Creates a blastdb from sequences grabbed from the input directory"""
        self.validate_files()

        if verbose:
            query_details = self._fasta_details(self.query_path)
            subject_details = self._fasta_details(self.subject_path)

            print("Making BLAST database from:")
            print(json.dumps(subject_details, indent=2))

            print("Query:")
            print(json.dumps(query_details, indent=2))

        self.run_cmd(
            "makeblastdb",
            dbtype="nucl",
            title=self.db_name,
            out=self.db,
            **{"in": self.subject_path}
        )
        return self.db

    def closedb(self):
        os.remove(self.results_out_path)
        shutil.rmtree(self.db_output_directory)

    @staticmethod
    def _filter_unique_results(results):
        return list(
            unique_everseen(
                results,
                key=lambda x: (
                    x["query"]["sequence_id"],
                    x["query"]["start"],
                    x["query"]["end"],
                    x["subject"]["sequence_id"],
                    x["subject"]["start"],
                    x["subject"]["end"],
                ),
            )
        )

    def _filter_results(self, results):
        results = self._filter_unique_results(results)
        return results

    def parse_results(self, delim=","):
        """
        Parses the raw blast result to a JSON

        :param delim: delimiter to parse
        :type delim: str
        :return:
        :rtype:
        """
        results = BlastResultParser.raw_results_to_json(self.raw_results, delim=delim)
        results = self._filter_results(results)
        self.results = results
        return self.results

    def get_perfect(self):
        return BlastResultParser.get_perfect(self.results)

    def __str__(self):
        return "{}".format(self.create_config())


class TmpBlast(BlastBase):
    """
    A Blast object that stores the database files in a hidden temporary directory. Use
    "quick_blastn" for returning results as a python object.
    """

    def __init__(self, db_name, subject_path, query_path, config=None):
        """
        TmpBlast: A Blast object that stores the database files in a hidden temporary directory.
        :param db_name: Name for database file structure. This name will be appended to all db files that blast creates.
        :param subject_path: Input directory or file containing a list of subjects to align against the query
        :param query_path: Location of the fasta or genbank file containing the query
        :param config: Additional configurations to run for the blast search (see
        https://www.ncbi.nlm.nih.gov/books/NBK279682/)
        """
        self.subject_records_path = subject_path

        super(TmpBlast, self).__init__(
            db_name=db_name,
            subject_path=None,
            query_path=query_path,
            db_output_directory=None,
            results_out_path=None,
            config=config,
        )

        self.fields = self.fields + (
            "subject_name",
            "subject_filename",
            "subject_circular",
            "query_name",
            "query_filename",
            "query_circular",
        )

    def makedb(self, **kwargs):
        db_output_directory = tempfile.mkdtemp()
        _, out = tempfile.mkstemp(dir=db_output_directory)

        if os.path.isdir(self.subject_records_path):
            self.subject_path = glob_fasta_to_tmpfile(self.subject_records_path)
        else:
            self.subject_path = self.subject_records_path
        self.set_paths(self.db_name, db_output_directory, out)
        super(TmpBlast, self).makedb(**kwargs)

    @classmethod
    def use_test_data(cls):
        """Create a Blast instance using predefined data located in tests"""
        dir_path = os.path.dirname(os.path.realpath(__file__))

        return cls(
            "db",
            os.path.join(dir_path, "..", "..", "tests/data/test_data/db.fsa"),
            os.path.join(dir_path, "..", "..", "tests/data/test_data/query.fsa"),
        )


class BioBlast(TmpBlast):

    DEFAULT_REINDEX = 1
    REINDEX_KEY = "reindex"

    def __init__(
        self,
        subjects: typing.List[SeqRecord],
        queries: typing.List[SeqRecord],
        seq_db=None,
        span_origin=True,
        force_unique_ids=False,
        config=None,
    ):
        """
        Initialize a new BioBlast.

        :param subjects: list of SeqRecords to use as subjects. Subjects get aligned to the queries.
        :type subjects: list
        :param queries: list of SeqRecords to use as queries. Subjects get aligned to the queries.
        :type queries: list
        :param seq_db: optional SeqRecordDB to use
        :type seq_db: SeqRecordDB
        :param span_origin: if False, will treat all sequences as linear (default: True)
        :type span_origin: bool
        :param force_unique_ids: if True, will force any records with the same record_id to be unique.
        :type force_unique_ids: bool
        :param config: additional blast config
        :type config: dict
        """
        if force_unique_ids:
            force_unique_record_ids(subjects)
            force_unique_record_ids(queries)
        self._check_records(subjects, queries)
        self.span_origin = span_origin
        if seq_db is None:
            self.seq_db = SeqRecordDB()
            _, subjects = self.add_records(
                subjects, seq_db=self.seq_db, span_origin=self.span_origin
            )
            _, queries = self.add_records(
                queries, seq_db=self.seq_db, span_origin=self.span_origin
            )
        else:
            self.seq_db = seq_db
        db_name = str(uuid4())
        subject_path = records_to_tmpfile(subjects)
        query_path = records_to_tmpfile(queries)
        super().__init__(
            db_name=db_name,
            subject_path=subject_path,
            query_path=query_path,
            config=config,
        )

        self.parse_options = {self.REINDEX_KEY: self.DEFAULT_REINDEX}

    def _check_records(self, subjects, queries):
        self._check_empty_records(queries, subjects)
        self._check_duplicate_records(queries, subjects)

    @staticmethod
    def _check_duplicate_records(queries, subjects):
        # Check SeqRecords have unique record_ids
        subject_rec_ids = {}
        for s in subjects:
            subject_rec_ids.setdefault(s.id, list()).append(s)
        query_rec_ids = {}
        for s in queries:
            query_rec_ids.setdefault(s.id, list()).append(s)
        duplicate_subjects = {k: v for k, v in subject_rec_ids.items() if len(v) > 1}
        duplicate_queries = {k: v for k, v in query_rec_ids.items() if len(v) > 1}
        suggestion = (
            "This may be due to clipping of the record_id that occurs when saving or loading"
            " certain file formats. Use `{}` or if using `pyblast.utils.load_glob` set the"
            "`force_unique_ids=True`.".format(force_unique_record_ids.__name__)
        )
        if duplicate_subjects:
            raise PyBlastException(
                "One or more subjects have the same record_id:\n{}.\n{}".format(
                    list(duplicate_subjects.keys()), suggestion
                )
            )
        if duplicate_queries:
            raise PyBlastException(
                "One or more queries have the smae record_id:\n{}\n{}".format(
                    list(duplicate_queries.keys()), suggestion
                )
            )

    @staticmethod
    def _check_empty_records(queries, subjects):
        # Check SeqRecords exist
        if not subjects:
            raise ValueError("Subjects is empty.")
        if not queries:
            raise ValueError("Queries is empty.")

    @staticmethod
    def _filter_remove_same_results(results):
        """
        Removes alignments that aligned to themselves.

        :param results:
        :type results:
        :return:
        :rtype:
        """
        new_results = []
        for r in results:
            query = r["query"]
            subj = r["subject"]
            k1 = query["origin_key"]
            k2 = subj["origin_key"]
            qends = (query["start"], query["end"])
            sends = (subj["start"], subj["end"])
            if k1 == k2 and qends == sends:
                continue
            else:
                new_results.append(r)
        return new_results

    @staticmethod
    def _filter_unique_results(results):
        return list(
            unique_everseen(
                results,
                key=lambda x: (
                    x["query"]["sequence_id"],
                    x["query"]["start"],
                    x["query"]["end"],
                    x["query"]["raw_end"],
                    x["subject"]["sequence_id"],
                    x["subject"]["start"],
                    x["subject"]["end"],
                    x["subject"]["raw_end"],
                ),
            )
        )

    def _filter_results(self, results):
        results = self._filter_unique_results(results)
        results = self._filter_remove_same_results(results)
        return results

    @classmethod
    def add_records(
        cls, records: typing.List[SeqRecord], seq_db: SeqRecordDB, span_origin=True
    ) -> typing.Tuple[typing.List[str], typing.List[SeqRecord]]:
        """
        Adds records to the local sequence database (SeqRecordDb). If `self.span_origin=True`,
        then in addition to adding the original SequenceRecord to the database, sequences will
        be pseudocircularized (concatenated with itself) if they are detected to be circular (via the 'topology'
        annotation') so that Blast can properly align sequences over the origin. Only sequences that were linear or
        the pseudociruclarized sequences are used in the alignment. After running blast and obtaining results, span
        indices are automatically corrected to account for the original pseudocircularization.

        :param records: list of SeqRecords annotated with 'topology' being 'linear' or 'circular'
        :type records: list
        :param seq_db: optional SeqRecordDB to use
        :type seq_db: SeqRecordDB
        :param span_origin: whether to search the sequence as a circular plasmid
        :type span_origin: bool
        :return: list of keys to use in the alignment procedure
        :rtype: list
        """
        clean_records(records)

        def copy_record(r):
            c = r[:]
            c.annotations = dict(r.annotations)
            return c

        def pseudocircularize(r):
            r2 = r + r
            r2.name = C.PSEUDOCIRCULAR + "__" + r.name
            r2.id = str(uuid4())
            return r2

        circular = [r for r in records if seq_db.is_circular(r)]
        linear = [r for r in records if not seq_db.is_circular(r)]
        if span_origin:
            keys = seq_db.add_many_with_transformations(
                circular, pseudocircularize, C.PSEUDOCIRCULAR
            )
        else:
            keys = seq_db.add_many_with_transformations(
                circular, copy_record, C.PSEUDOCIRCULAR
            )
        keys += seq_db.add_many_with_transformations(linear, copy_record, C.COPY_RECORD)
        transformed_records = seq_db.get_many(keys)
        return keys, transformed_records

    def parse_result_to_span(
        self, data, inclusive=True, input_index=1, output_index=None
    ):

        if output_index is None:
            output_index = input_index

        s, e, l = data["start"], data["raw_end"], data["origin_sequence_length"]
        if data["strand"] == -1:
            s, e = e, s
        c = data["circular"]
        if inclusive:
            e += 1
        span = Span(s, e, l, cyclic=c, allow_wrap=True, index=input_index)
        if input_index != output_index:
            span = span.reindex(output_index)
        return span

    @classmethod
    def parse_to_span(cls, v, reindex=0):
        """Convert a JSON blast result to a Span. Optionally reindex"""

        def make_span(data):
            cls.parse_result_to_span(
                data=data,
                inclusive=v["meta"]["inclusive"],
                input_index=v["meta"]["start_index"],
                output_index=reindex,
            )

        return {"query": make_span(v["query"]), "subject": make_span(v["subject"])}

    def _parse__annotate_meta(self, meta):
        """
        Annotate default meta information from the raw JSON result.

        :param meta:
        :return:
        """
        meta["start_index"] = 1
        meta["inclusive"] = True
        meta["span_origin"] = self.span_origin

    def _parse__annotate_record(self, data):
        """
        Annotate record information from the raw JSON result using the sequence database and sequence_id.

        :param data:
        :return:
        """
        origin_key = self.seq_db.get_origin_key(data["sequence_id"])
        record = self.seq_db.get(origin_key)
        is_circular = self.seq_db.is_circular(record)
        data["circular"] = is_circular
        data["name"] = record.name
        data["origin_key"] = origin_key
        data["origin_record_id"] = record.id
        data["origin_sequence_length"] = len(record.seq)

    def _parse__correct_endpoints(
        self, data, inclusive=True, input_index=1, output_index=None
    ):
        """
        Correct endpoints of JSON results.

        :param data: The JSON result. E.g. `{"start": 0, "end": 100, "origin_sequence_length": 200}`
        :param inclusive: If True, the input JSON result is considered to be inclusive at the endpoint.
        :param input_index: The starting index assumed in the input JSON result.
        :param output_index: The index to reindex the result.
        :return:
        """

        if output_index is None:
            output_index = input_index
        data["raw_end"] = data["end"]
        s, e = data["start"], data["end"]
        reverse = data["strand"] != 1
        if reverse:
            s, e = e, s
        if e - s < 0:
            raise ValueError("End cannot be less than start")
        elif data["circular"]:
            span = self.parse_result_to_span(
                data=data,
                inclusive=inclusive,
                input_index=input_index,
                output_index=output_index,
            )
            data["start"] = span.a
            data["end"] = span.b - 1
            data["raw_end"] = span.c - 1
            if reverse:
                data["start"], data["end"], data["raw_end"] = (
                    data["end"],
                    data["start"],
                    data["start"],
                )

    def _parse_bioblast_results(self, v, output_index):
        self._parse__annotate_meta(v["meta"])
        for x in ["query", "subject"]:
            self._parse__annotate_record(v[x])
            self._parse__correct_endpoints(
                v[x],
                inclusive=v["meta"]["inclusive"],
                input_index=v["meta"]["start_index"],
                output_index=output_index,
            )
        v["meta"]["start_index"] = output_index
        return v

    def parse_results(self, delim=",", reindex=None):
        """
        Parses the blast output to a digestable JSON output of the following format. Results
        can be found in `self.results` or returned from this function. Starting and ending
        positions are inclusive and starting index is 1.

        ..code-block::

            [
                {
                  "query": {
                    "start": 1,
                    "end": 4219,
                    "bases": "TTTAGTATATAT...",
                    "strand": 1,
                    "length": 18816,
                    "sequence_id": "pMODKan-HO-pACT1-Z4-"
                  },
                  "subject": {
                    "start": 1,
                    "end": 4219,
                    "bases": "TCGCGCGTTTCGGTGATGACGG...",
                    "strand": 1,
                    "length": 7883,
                    "sequence_id": "pMODKan-HO-pACT1-ZEV4"
                  },
                  "meta": {
                    "query acc.": "pMODKan-HO-pACT1-Z4-",
                    "subject acc.": "pMODKan-HO-pACT1-ZEV4",
                    "score": 4219,
                    "evalue": 0,
                    "bit score": 7792,
                    "alignment length": 4219,
                    "identical": 4219,
                    "gap opens": 0,
                    "gaps": 0,
                    "query length": 18816,
                    "q. start": 1,
                    "q. end": 4219,
                    "subject length": 7883,
                    "s. start": 1,
                    "s. end": 4219,
                    "subject strand": "plus",
                    "query seq": "TTTAGTATATAT...",
                    "subject seq": "TCGCGCGTTTCGGTGATGACGG..."
                  }
                }
            ]

        :param delim:
        :type delim:
        :return:
        :rtype:
        """
        if reindex is None:
            reindex = self.parse_options.get(self.REINDEX_KEY, self.DEFAULT_REINDEX)
        parsed_results = BlastResultParser.raw_results_to_json(
            self.raw_results, delim=delim
        )

        # # TODO: resolve with sequence dictionary, resolving pseudocircularized constructs
        for v in parsed_results:
            if v:
                self._parse_bioblast_results(v, output_index=reindex)
        parsed_results = self._filter_results(parsed_results)
        self.results = parsed_results
        return self.results


class JSONBlast(BioBlast):
    """Object that runs blast starting from JSON inputs and outputs"""

    def __init__(
        self, subject_json, query_json, span_origin=True, seq_db=None, config=None
    ):
        """
        Initialize JSONBlast

        :param subject_json: subject sequences as serialized json. Schema may be found in :class:`pyblast.schema.SequenceSchema`
        :type subject_json: list of dict, mapping, or Object
        :param query_json: query sequence as serialized json
        :type query_json: dict, mapping, or Object
        :param span_origin: default False. If True, circular sequences will be pseudocircularized for alignment over the origin
        :type span_origin: boolean
        :param seq_db: optional SeqRecordDB to use
        :type seq_db: SeqRecordDB
        :param config: optional arguments
        :type config: dict
        """
        srecords = JSONParser.JSON_to_SeqRecords(subject_json)
        qrecords = JSONParser.JSON_to_SeqRecords(query_json)
        self.subject_json = subject_json
        self.query_json = query_json
        super().__init__(
            srecords, qrecords, seq_db=seq_db, span_origin=span_origin, config=config
        )

    @classmethod
    def use_test_data(cls):
        """Create a Blast instance using predefined data located in tests"""
        dir_path = os.path.dirname(os.path.realpath(__file__))
        with open(
            os.path.join(dir_path, "..", "..", "tests/data/test_data/templates.json"),
            "r",
        ) as f:
            subject = json.load(f)
        with open(
            os.path.join(dir_path, "..", "..", "tests/data/test_data/query.json"), "r"
        ) as f:
            query = json.load(f)
        return cls(subject_json=subject, query_json=query)



# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['pyblast', 'pyblast.blast', 'pyblast.blast_bin', 'pyblast.utils']

package_data = \
{'': ['*']}

install_requires = \
['biopython>=1.73,<2.0',
 'fire>=0.1.3,<0.2.0',
 'loggable-jdv>=0.1.4,<0.2.0',
 'networkx>=2.3,<3.0',
 'termcolor>=1.1,<2.0']

entry_points = \
{'console_scripts': ['pyblast = pyblast:cli.main']}

setup_kwargs = {
    'name': 'pyblastbio',
    'version': '0.4.1',
    'description': '',
    'long_description': '[![travis build](https://img.shields.io/travis/jvrana/pyblast.svg)](https://travis-ci.org/jvrana/pyblast)\n[![Coverage Status](https://coveralls.io/repos/github/jvrana/pyblast/badge.svg?branch=master)](https://coveralls.io/github/jvrana/pyblast?branch=master)\n[![PyPI version](https://badge.fury.io/py/REPO.svg)](https://badge.fury.io/py/REPO)\n\n![module_icon](images/module_icon.png?raw=true)\n\n#### Build/Coverage Status\n|     Branch      |                            Build                             |                           Coverage                           |\n| :-------------: | :----------------------------------------------------------: | :----------------------------------------------------------: |\n|   **master**    | [![travis build](https://img.shields.io/travis/jvrana/pyblast/master.svg)](https://travis-ci.org/jvrana/pyblast/master) | [![Coverage Status](https://coveralls.io/repos/github/jvrana/pyblast/badge.svg?branch=master)](https://coveralls.io/github/jvrana/pyblast?branch=master) |\n| **development** | [![travis build](https://img.shields.io/travis/jvrana/pyblast/development.svg)](https://travis-ci.org/jvrana/pyblast/development) | [![Coverage Status](https://coveralls.io/repos/github/jvrana/pyblast/badge.svg?branch=development)](https://coveralls.io/github/jvrana/pyblast?branch=development) |\n\n# pyblast\n\nThis is a wrapper for other applications to run blast searches on SeqRecord objects and JSON objects. Intended to\nbe used in small python applications.\n\nFeatures include:\n* Automatic BLAST parsing to JSON\n* Alignment to circular queries, using either linear or circular subjects\n* Blast self installation\n\n# Installation\n\nYou can install BLAST to the pyblast directory using the following command:\n\n```\npyblast install\n```\n\nThis will install it to pyblast/blast_bin in your python install location. If you want BLAST installed somewhere else, move the *ncbi-blast-X.X.X+* folder\nto your desired location and add *path/to/ncbi-blast-X.X.X+/bin* to you $PATH. **PyBlast** will prefer to use the blast stored\nin your executable path. If it cannot find a blast executable there, it looks for it in that paths in the pyblast/blast_bin/_paths.txt.\nfile. _paths.txt is automatically updated when you run install_blast.py so theres no need to manage the paths manually.\n\nAfter installing and verifying the `blastn` command works from the cmd line,\n\n```\npip install pyblastbio\n```\n\n## Usage\n\nThis package is a python wrapper for the BLAST command line, intended to be run along with a microservice (e.g. Flask) or for a quick alignment in a jupyter notebook or small python script/app.\n\nThis package also includes a basic python-based installation script which is used in unit-testing.\n\n### Running a blast query on a Bio.SeqRecord object\n\nWe can do a quick alignment to some sequences using the following, which gives us a nice dictionary of the results:\n\n```python\nfrom pyblast import BioBlast\nfrom pyblast.utils import make_linear, make_circular\nfrom Bio.SeqRecord import SeqRecord\nfrom Bio.Seq import Seq\n\nqueries = [\n  SeqRecord(Seq("ACGTGATTCGTCGTGTAGTTGAGTGTTACGTTGCATGTCGTACGTGTGTAGTGTCGTGTAGTGCTGATGCTACGTGATCG"))\n]\nsubjects = [\n  SeqRecord(Seq("ACGTGATTCGTCGTGTAGTTGAGTGTTACGTTGCATGTCGTTACGTGATCG"))\n]\n\n# pyblast requires a \'topology\' annotation on the SeqRecords.\n# we can make records circular or linear using `make_linear` or `make_circular` methods\nsubjects = make_linear(subjects)\nqueries = make_linear(queries)\n\nblast = BioBlast(subjects, queries)\nresults = blast.quick_blastn()\nprint(results)\n```\n\n```json\n[\n  {\n    "query": {\n      "start": 1,\n      "end": 46,\n      "bases": "ACGTGATTCGTCGTGTAGTTGAGTGTTACGTTGCATGTCGT-ACGTG",\n      "strand": 1,\n      "length": 80,\n      "sequence_id": "11e17df2-579f-4234-a1e6-f4e3fadfe277",\n      "circular": false,\n      "name": "<unknown name>",\n      "origin_key": "bbadd55c-9413-4394-a23c-0da983630b98",\n      "origin_record_id": "<unknown id>",\n      "origin_sequence_length": 80\n    },\n    "subject": {\n      "start": 1,\n      "end": 47,\n      "bases": "ACGTGATTCGTCGTGTAGTTGAGTGTTACGTTGCATGTCGTTACGTG",\n      "strand": 1,\n      "length": 51,\n      "sequence_id": "69248d23-1044-4a75-80c9-53b999796d48",\n      "circular": false,\n      "name": "<unknown name>",\n      "origin_key": "1f627d51-93df-458b-ba36-9b5a7b483a4d",\n      "origin_record_id": "<unknown id>",\n      "origin_sequence_length": 51\n    },\n    "meta": {\n      "query acc.": "11e17df2-579f-4234-a1e6-f4e3fadfe277",\n      "subject acc.": "69248d23-1044-4a75-80c9-53b999796d48",\n      "score": 43,\n      "evalue": 0,\n      "bit score": 80,\n      "alignment length": 47,\n      "identical": 46,\n      "gap opens": 1,\n      "gaps": 1,\n      "query length": 80,\n      "q. start": 1,\n      "q. end": 46,\n      "subject length": 51,\n      "s. start": 1,\n      "s. end": 47,\n      "subject strand": "plus",\n      "query seq": "ACGTGATTCGTCGTGTAGTTGAGTGTTACGTTGCATGTCGT-ACGTG",\n      "subject seq": "ACGTGATTCGTCGTGTAGTTGAGTGTTACGTTGCATGTCGTTACGTG",\n      "span_origin": true\n    }\n  }\n]\n```\n\n### Running blast on circular subjects and queries\n\nPyblast handles alignments to circular subjects and queries as well. As you can see below, we get a complete alignment of the subject (1 to 50) to the circular query (82 over origin to 30). Circular subjects and circular queries can be mixed together, as well as multiple queries.\n\n```\nseq = "ACGTTGTAGTGTAGTTGATGATGATGTCTGTGTCGTGTGATGTGCTGTAGTGTTTAGGGGCGGCGCGGAGTATGCTG"\nqueries = [\n\tSeqRecord(Seq(seq))\n]\n\nsubjects = [\n\tSeqRecord(Seq(seq[-20:] + seq[:30]))\n]\n\n# pyblast requires a \'topology\' annotation on the SeqRecords.\n# we can make records circular or linear using `make_linear` or `make_circular` methods\nsubjects = make_circular(subjects)\nqueries = make_circular(queries)\n\nblast = BioBlast(subjects, queries)\nresults = blast.quick_blastn()\nprint(results)\n```\n\n```json\n[\n  {\n    "query": {\n      "start": 82,\n      "end": 30,\n      "strand": 1,\n      "...": "..."\n    },\n    "subject": {\n      "start": 1,\n      "end": 50,\n      "strand": 1,\n      "...": "..."\n    },\n    "meta": {\n    \t"...": "..."\n    }\n]\n```\n\n### BioBlastFactory\n\nIn some cases, we will want to share the same sequences for different types of alignments. For example, we may want to align a set of primers and a set of templates to the same query records. In these types of cases, we can use the **BioBlastFactory**:\n\n```python\nfrom pyblast import BioBlastFactory\n\n# initialize a new factory\nfactory = BioBlastFactory()\n\n# add records accessible by keyword\nfactory.add_records(records1, "primers")\nfactory.add_records(records2, "templates")\nfactory.add_records(records3, "queries")\n\n# we spawn new BioBlast alignmers from the keywords above\nprimer_alignment = factory("primers", "queries")\ntemplate_alignment = factory("templates", "queries")\n\n# we can then run alignments, ensuring the queries in both results\n# refer to the exact same query\nprimer_results = primer_alignment.quick_short_blastn()\ntemplate_results = template_alignment.quick_blastn()\n```\n\n### Utilities for reading files\n\n**pyblast** includes utilities for reading in *fasta* and *genbank* files.\n\n```python\nfrom pyblast.utils import load_glob, load_genbank_glob, load_fasta_glob\n\n# load many genbank files into a list of SeqRecords\n# \'topology\' is automatically detected here\n# we enforce all record_ids to be unique (a requirement for pyblast)\nrecords1 = load_genbank_glob("~/mydesigns/*.gb", force_unique_ids=True)\n\n# load many fasta files into a list of SeqRecords\n# \'topology\' is NOT detected\n# we enforce all record_ids to be unique (a requirement for pyblast)\nrecords2 = make_linear(load_fasta_glob("~/mydesigns/*.fasta"), force_unique_ids=True)\n\n```',
    'author': 'Justin Vrana',
    'author_email': 'justin.vrana@gmail.com',
    'url': 'https://www.github.com/jvrana/pyblast',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



from __future__ import annotations
from collections.abc import Container, Iterable, Sized
from itertools import chain
from typing import List, Tuple, Union


class SpanError(Exception):
    pass


class Span(Container, Iterable, Sized):
    __slots__ = ["_a", "_b", "_c", "_context_length", "_cyclic", "_index", "_strict"]

    def __init__(
        self,
        a: int,
        b: int,
        l: int,
        cyclic=False,
        index=0,
        allow_wrap=True,
        strict=False,
    ):
        """
        Constructs a new Span. `Span` maps the provided positions onto a context
        Spans have no direction and have an underlying context
        that has a certain length, can be linear or cyclic, and has a starting index.

        The following attributes are accessible:

        - `a` - the starting (inclusive) position of the span
        - `b` - the mapped exclusive position of the span
        - `c` - the unmapped exclusive position of the span (used for cyclic spans that wrap context multiple times)
        - `cyclic` - whether the underlying context is circular or linear
        - `index` - the first index of the context
        - `context_length` - the length of the underlying context.

        **Linear spans**
        For example, a basic linear span can be represented using the following:

        .. code-block::

            s = Span(0, 10, 20)
            assert s.a == 0
            assert s.b == 10
            assert list(s) == [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]  # spans are always exclusive at endpoint.

        **Indexing and Slicing**

        New indexes can be provided to spans.

        .. code-block::

            s = Span(1, 10, 10, index=1)
            assert s.a == 1
            assert s.b == 10

        Positions can be mapped to the span:

        .. code-block::

            s = Span(5, 8, 10, index=5)
            assert s.a == 5
            assert s.b == 8
            assert s[0] == 5  # the 'first' position in the span equivalent to the starting index
            assert s[-1] == 7  # the 'last' position in the span is the last inclusive index

        Positions can be mapped automatically during initialization:

        .. code-block::

            s = Span(-1, 5, 10, cyclic=True, index=0)  # index '-1' is mapped onto last available index on the context,
            '9'
            assert s.a == 9
            assert s.b == 5

            s = Span(6, -1, 10, cyclic=True, index=0)  # index '-1' is mapped onto last available exclusive index on the
             context, '10'
            assert s.a == 6
            assert s.b == 10

        Inclusive positions can be mapped onto the context using `t`:

        .. code-block::

            s = Span(-1, 5, 10, cyclic=True)
            assert s.t(11) == 1
            assert s.t(10) == 0

        **Reindexing**

        Context starting index can be remapped:

        .. code-block::

            s = Span(0, 5, 10)
            s1 = s.reindex(1)
            asset s1.a == 1
            assert s2.b == 6

        **Slicing**

        Spans can be sliced similarly to lists:

        .. code-block::

            s = Span(4, 8, 10)
            s1 = s[1:]
            assert s1.a == 5
            assert s2.b == 8

            s2 = s[:-1]
            assert s2.a == 4
            assert s2.b == 7

            s3 = s[2:-1]
            assert s3.a == 5
            assert s3.b == 7

        **Cyclic spans**
        A cyclic span can be represented in the following way:

        .. code-block::

            s = Span(18, 2, 20, cyclic=True)
            assert s.a == 18
            assert s.b == 2
            assert list(s) == [18, 19, 0, 1]
            assert len(s) == 4

        **Wrapping cyclic spans**
        Spans the wrap around the context multiple times can be represented as well. The mapped endpoint
        is found with `span.b` and the non-mapped endpoint is found using `span.c`.

        .. code-block::

            s = Span(9, 21, 10, cyclic=True)
            assert list(s) == [9, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 0]
            assert len(s) == 12
            assert s.a == 9  # start point
            assert s.b == 1  # mapped endpoint
            assert s.c == 21 # the unmapped endpoint

        The indices are reduced to their lowest 'wrapping' whenever possible. For example, the following
        initializations are equivalent:

        .. code-block::

            s1 = Span(1, 5, 10, cyclic=True)
            s2 = Span(11, 15, 10, cyclic=True)
            s3 = Span(21, 25, 10, cyclic=True)
            assert s1 == s2
            assert s2 == s3

        **Iterating through spans**

        Spans can be treated like iterators:

        .. code-block::

            for i in Span(5, 3, 10, cyclic=True):
                print(i)

        **Span inversion**

        Inversion returns two spans that represent everything *except* the span:

        .. code-block::
            s = Span(4, 8, 10)
            s1, s2 = s.invert()  # also s[::-1]
            assert s1.a == 0
            assert s1.b == 4
            assert s2.a == 8
            assert s2.b == 10

            # inverted cyclic span returns one valid span and None
            s = Span(4, 8, 10, cyclic=True)
            s1, s2 = s.invert()
            assert s1.a == 8
            assert s1.b == 4
            assert s2 is None

        **Intersections, differences**

        .. code-block::

            s1 = Span(4, 10, 20)
            s2 = Span(8, 12, 20)
            s3 = s1.intersection(s2)
            assert s3.a == 8
            assert s3.b == 10

        .. code-block::

            s1 = Span(4, 10, 20)
            s2 = Span(8, 12, 20)
            s3, s4 = s1.differences(s2)

            assert s3.a, s3.b == 4, 8
            assert s4,a, s4.b == 10, 12

        :param a: start of the span (inclusive)
        :type a: int
        :param b: end of the span (exclusive)
        :type b: int
        :param l: context length of the region
        :type l: int
        :param cyclic: whether the underlying context is cyclic
        :type cyclic: bool
        :param index: the starting index of the region
        :type index: int
        :param strict: if True, positions outside of context bounds are disallowed.
        :type strict: bool
        :param allow_wrap: if True (default False), spans that wrap around context more than once will be mapped
                            as if the span only wrapped around once.
        :type allow_wrap: bool
        """

        self._context_length = l
        self._index = index
        self._cyclic = cyclic
        self._strict = strict

        # special empty edge case
        if cyclic and a == b:
            self._set_as_empty(a)
            return

        # check bounds
        if self._strict or not cyclic:
            bounds = self.bounds()
            if not bounds[0] <= a < bounds[1]:
                raise IndexError(
                    "Start {} must be in [{}, {})".format(a, index, index + l)
                )
            if not bounds[0] <= b <= bounds[1]:
                raise IndexError(
                    "End {} must be in [{}, {}]".format(b, index, index + l)
                )

        start_wrap = int((a - index) / l)
        end_wrap = int((b - index - 1) / (l))

        if start_wrap > end_wrap:
            self._a = a
            self._b = b
            self._c = b
            diff = end_wrap - start_wrap
            # return self._set_as_empty(a)
            raise IndexError(
                "Could not interpret span {span}. Starting position wraps around "
                "context {i} times and end position wraps around {j} times."
                " A valid initialization would be Span({a}, {b}, ...)".format(
                    span=self,
                    i=start_wrap,
                    j=end_wrap,
                    a=self._a,
                    b=self._b - diff * self._context_length,
                )
            )

        # set indices
        _a = a - index
        _b = b - index

        if _a >= l or _a < 0:
            self._a = self.t(_a, False)
        else:
            self._a = a
        if _b > l:
            self._b = self.t(_b - 1, False) + 1
        elif _b < 0:
            self._b = self.t(_b, False) + 1
        else:
            self._b = b

        if self._a > self._b and not cyclic:
            raise IndexError(
                "Start {} cannot be greater than end {} for linear spans.".format(
                    self._a, self._b
                )
            )

        # allow wrap mean this will keep track of how many time the span wraps around the context
        if allow_wrap and end_wrap - start_wrap:
            _c = self._b + (end_wrap - start_wrap) * l
            self._c = self._b + (end_wrap - start_wrap) * l
        else:
            self._c = self._b

    @property
    def index(self):
        """Return the starting index of the context."""
        return self._index

    @property
    def a(self):
        """Return the inclusive mapped startpoint."""
        return self._a

    @property
    def b(self):
        """Return the exclusive mapped endpoint."""
        return self._b

    @property
    def c(self):
        """Return the exclusive un-mapped endpoint."""
        return self._c

    @property
    def cyclic(self):
        """Return whether the context is cyclic/circular."""
        return self._cyclic

    @property
    def context_length(self):
        """Return the length of the context."""
        return self._context_length

    def _set_as_empty(self, a: int) -> None:
        _a = self.t(a - self._index, False)
        self._a = self._b = self._c = _a
        return

    @property
    def _nwraps(self):
        return int((self._c - self._index - 1) / (self._context_length))

    def bounds(self) -> tuple:
        """Return the context bounds (end exclusive)"""
        return self._index, self._context_length + self._index

    def t(self, p: int, throw_error=True) -> int:
        """
        Translates a position 'p' to an index within the context bounds.

        :param p:
        :type p:
        :param throw_error:
        :return:
        :rtype:
        """
        if p >= self.bounds()[1] or p < self.bounds()[0]:
            if throw_error and not self._cyclic:
                raise IndexError(
                    "Position {} outside of linear bounds {}".format(p, self.bounds())
                )
        _x = p % self._context_length
        if _x < 0:
            return self.bounds()[1] + _x
        else:
            return self.bounds()[0] + _x

    @staticmethod
    def _ranges_str(ranges):
        s = ",".join("[{}, {})".format(*r) for r in ranges)
        return "[" + s + "]"

    def ranges(self) -> List[Tuple[int, int]]:
        """
        Return ranges of valid positions.

        :return:
        :rtype:
        """

        # TODO: replace this with _nwraps > 0
        if self._cyclic and (self._b < self._a or self._nwraps):
            ranges = [(self._a, self.bounds()[1])]
            for _ in range(self._nwraps - 1):
                ranges.append(self.bounds())
            ranges.append((self.bounds()[0], self._b))

            return ranges
        else:
            return [(self._a, self._b)]

    def slices(self):
        """
        Return list of slices

        :return:
        :rtype:
        """
        return [slice(*r) for r in self.ranges()]

    def reindex(self, i, strict=None, allow_wrap=True):
        """
        Return a new span with positions reindexed.

        :param i: new index
        :param strict: initialize with 'strict'
        :param allow_wrap: allow multiple wrapping.
        :return:
        """
        return self.new(None, None, allow_wrap=allow_wrap, index=i, strict=strict)

    def new(
        self,
        a: Union[int, None],
        b: Union[int, None],
        allow_wrap=True,
        index=None,
        strict=None,
    ) -> Span:
        """Create a new span using the same context."""
        if a is None:
            a = self._a
        if b is None:
            b = self._c

        if strict is None:
            strict = self._strict

        if index is not None:
            d = index - self._index
            a += d
            b += d
        else:
            index = self._index
        return self.__class__(
            a,
            b,
            self._context_length,
            self._cyclic,
            index=index,
            allow_wrap=allow_wrap,
            strict=strict,
        )

    def sub(self, a: int, b: int) -> Span:
        """
        Create a sub region starting from a to b.

        :param a: starting pos
        :param b: ending pos
        :return: a new Span.
        """
        # if a == b:
        #     return self.new(a, b)
        if b is not None and a > b and not self._cyclic:
            raise ValueError(
                "Start {} cannot exceed end {} for linear spans".format(a, b)
            )

        if self._nwraps > 0:
            valid_ranges = [(self._a, self._c)]
        else:
            valid_ranges = [list(x) for x in self.ranges()]

            valid_ranges[0][0] = max(a, self._a)
            valid_ranges[-1][1] = min(b + 1, self._b + 1)
        assert len(valid_ranges) <= 2

        def in_range(pos, ranges):
            for i, r in enumerate(ranges):
                if r[0] <= pos < r[1]:
                    return (True, i)
            return (False, None)

        start_in_range, start_range = in_range(a, valid_ranges)

        if not start_in_range:
            raise IndexError(
                "Start {} must be in {}".format(a, self._ranges_str(valid_ranges))
            )

        valid_end_range = valid_ranges[start_range:]
        end_in_range, end_range = in_range(b, valid_end_range)

        if not end_in_range:
            raise IndexError(
                "End {} must be in {}".format(b, self._ranges_str(valid_end_range))
            )
        subregion = self.new(a, b)
        if len(subregion) > len(self):
            raise IndexError("Cannot make subspan.")
        return subregion

    def same_context(self, other: Span) -> bool:
        """Return if another Span as an equivalent context."""
        return (
            other.context_length == self._context_length
            and self._cyclic == other.cyclic
        )

    def force_context(self, other: Span) -> None:
        """
        Raise error if another Span has different context.

        :param other: The other span
        :return: None
        :raises: SpanError
        """
        if not self.same_context(other):
            raise SpanError("Cannot compare with different contexts")

    def overlaps_with(self, other: Span) -> bool:
        """
        Returns True if other span has an overlap with this span.

        :param other: The other span
        :return: True if the two spans overlap.
        """
        self.force_context(other)
        # if other in self:
        #     return True
        # elif self in other:
        #     return True
        if (
            other.a in self
            or other.b - 1 in self
            or self._a in other
            or self._b - 1 in other
        ):
            return True
        return False

    def differences(self, other: Span) -> Union[Tuple[Span], Tuple[Span, Span]]:
        """Return a tuple of differences between this span and the other span."""
        self.force_context(other)
        if other.a in self and other.b not in self:
            return (self.new(self._a, other.a),)
        elif other.b in self and other.a not in self:
            return (self.new(other.b, self._b),)
        if other in self:
            return self.new(self._a, other.a), self.new(other.b, self._b)
        elif self in other:
            return (self.new(self._a, self._a),)
        else:
            return ((self[:]),)

    def intersection(self, other: Span) -> Span:
        """Return the span inersection between this span and the other span."""
        self.force_context(other)
        if other.a in self and other.b not in self:
            return self.sub(other.a, self._b)
        elif other.b in self and other.a not in self:
            return self.sub(self._a, other.b)
        if other in self:
            return other[:]
        elif self in other:
            return self[:]

    def consecutive(self, other: Span) -> bool:
        """Returns True if other span is immediately consecutive with this span."""
        self.force_context(other)
        try:
            return self._b == other.t(other.a)
        except IndexError:
            return False

    def connecting_span(self, other: Span) -> Union[Span, None]:
        """
        Return the span that connects the two spans. Returns None

        :param other:
        :return:
        """
        self.force_context(other)
        if self._cyclic and self._a == other.a and self._b == other.b:
            return self.invert()[0]
        if self.consecutive(other):
            return self[self._b, self._b]
        elif self.overlaps_with(other):
            return None
        else:
            if self._b > other.a and not self._cyclic:
                return None
            return self[self._b, other.a]

    # def union(self, other):
    #     self.force_context(other)
    #     if other in self:
    #         return self[:]
    #     elif self in other:
    #         return other[:]
    #     elif other.a in self and not other.t(other.b - 1) in self:
    #         if self._a == other.b:
    #             return self.new(self._a, None)
    #         return self.new(self._a, other.b)
    #     elif other.t(other.b - 1) in self and other.a not in self:
    #         if other.a == self._b:
    #             return self.new(self._a, None)
    #         return self.new(other.a, self._b)

    # def __ge__(self, other):
    #     self.force_context(other)
    #     return self._a >= other.a

    #     def __lt__(self, other):
    #         return self._a < other.a

    #     def __gt__(self, other):
    #         return self._a > other.a

    #     def __ge__(self, other):
    #         return self._a >= other.a

    # def __invert__(self):
    #     if self._a > self._b:
    #         if self._cyclic:
    #             return self[self._b+1, self._a-1],
    #         else:
    # return

    def invert(self) -> Union[Tuple[Span, Span], Tuple[Span, None]]:
        """
        Invert the region, returning a tuple of the remaining spans from the context.
        If cyclic, a tuple (span, None) tuple is returned. If linear, a (span, span) is returned.

        :return: inverted regions
        :rtype: tuple
        """
        if len(self) >= self._context_length:
            return (self[self._a, self._a], None)
        if self._cyclic:
            return (self[self._b, self._a], None)
        else:
            return self[:, self._a], self[self._b, :]

    def __eq__(self, other: Span) -> bool:
        return self.same_context(other) and self._a == other.a and self._b == other.b

    def __ne__(self, other: Span) -> bool:
        return not (self.__eq__(other))

    @classmethod
    def _pos_in_ranges(cls, pos: int, ranges: List[Tuple[int, int]]):
        for r in ranges:
            if r[0] <= pos < r[1]:
                return True
        return False

    def contains_pos(self, pos: int) -> bool:
        return self._pos_in_ranges(pos, self.ranges())

    def contains_span(self, other: Span) -> bool:
        if not self.same_context(other):
            return False
        if other.a == other.b == self._a:
            # special case where starting indices and ending indices are the same
            return True
        else:
            if not self.contains_pos(other.a):
                return False
            elif not self.contains_pos(other.t(other.b - 1)):
                return False
            elif not len(other) <= len(self):
                return False
            return True

    def spans_origin(self) -> bool:
        if self._nwraps and self._cyclic:
            return True
        return self._b < self._a and self._cyclic

    def __contains__(self, other: Union[Span, int]):
        if isinstance(other, int):
            return self.contains_pos(other)
        elif issubclass(type(other), Span):
            return self.contains_span(other)

    def __len__(self):
        return sum([r[1] - r[0] for r in self.ranges()])

    def __iter__(self):
        for i in chain(*[range(*x) for x in self.ranges()]):
            yield i

    def __invert__(self):
        return self.invert()

    def _check_index_pos(self, val: int) -> None:
        if val is not None and (val >= len(self) or val < -len(self)):
            raise IndexError(
                "Index '{}' outside of linear span with length of {}".format(
                    val, len(self)
                )
            )

    def __getitem__(self, val):
        if isinstance(val, int):
            self._check_index_pos(val)
            if val < 0:
                return self.t(val + self._c - self._index)
            else:
                return self.t(val + self._a - self._index)
        elif issubclass(type(val), slice):

            self._check_index_pos(val.start)
            self._check_index_pos(val.stop)

            if val.step == -1:
                return self[val.start : val.stop].invert()
            elif val.step is not None and val.step != 1:
                raise ValueError(
                    "{} slicing does not support step {}.".format(
                        self.__class__.__name__, val.step
                    )
                )

            if val.start is None:
                i = self._a
            elif val.start < 0:
                i = self._c + val.start
            else:
                i = self._a + val.start

            if val.stop is None:
                j = self._c
            elif val.stop < 0:
                j = self._c + val.stop
            else:
                j = self._a + val.stop

            return self.new(i, j)
        elif isinstance(val, tuple):
            if len(val) > 2:
                raise ValueError(
                    "{} -- copying only supports (start, stop)".format(val)
                )
            val = list(val)
            for i in [0, 1]:
                if val[i] == slice(None, None, None):
                    val[i] = None
            if val == (None, None):
                val = self.bounds()
            elif val[0] is None:
                val = (self.bounds()[0], val[1])
            elif val[1] is None:
                val = (val[0], self.bounds()[1])
            return self.new(*val)
        else:
            raise ValueError("indexing does not support {}".format(type(val)))

    def __repr__(self):
        return "<Span {a} {b} ({c}) len={l} cyclic={cyclic} index={index}, nwraps={n}>".format(
            self.__class__.__name__,
            a=self._a,
            b=self._b,
            c=self._c,
            l=self._context_length,
            cyclic=self._cyclic,
            index=self._index,
            n=self._nwraps,
        )

    def __str__(self):
        return self.__repr__()


class EmptySpan(Span):
    def ranges(self):
        return [(self.bounds()[0], self.bounds()[0])]



from .cmd import run_cmd
from Bio.Alphabet import generic_dna
from Bio.Seq import Seq
from glob import glob
from os.path import join
from Bio import SeqIO
from Bio.SeqRecord import SeqRecord
import tempfile
from Bio.SeqFeature import FeatureLocation, CompoundLocation
from pyblast.constants import Constants as C
from typing import List
from uuid import uuid4
from .span import Span, SpanError


def reverse_complement(seq_str: str) -> str:
    """
    Return reverse complement of nucleotide sequence

    :param seq_str:
    :type seq_str:
    :return:
    :rtype:
    """
    return str(Seq(seq_str, generic_dna).reverse_complement())


def complement(seq_str: str) -> str:
    """
    Return the complement of the nucleotide sequence.

    :param seq_str:
    :type seq_str:
    :return:
    :rtype:
    """
    return str(Seq(seq_str, generic_dna).complement())


def new_feature_location(start: int, end: int, length: int, strand: int):
    """Makes a FeatureLocation. If necessary, makes CompoundLocation."""
    if start > end:
        if length is None:
            raise ValueError(
                "A length must be provided to create a feature with start > end."
            )
        f1 = FeatureLocation(start, length, strand)
        f2 = FeatureLocation(1, end, strand)
        if strand == -1:
            location = CompoundLocation([f2, f1])
        else:
            location = CompoundLocation([f1, f2])
    else:
        location = FeatureLocation(start, end, strand=strand)
    return location


def records_to_tmpfile(records: List[SeqRecord]) -> str:
    """
    Write SeqRecords to a temporary file.

    :param records: list of SeqRecords
    :type records: list
    :return: temporary file path
    :rtype: str
    """
    fd, tmp_path_handle = tempfile.mkstemp(suffix=".fasta")
    SeqIO.write(records, tmp_path_handle, format="fasta")
    return tmp_path_handle


def glob_fasta_to_tmpfile(dirpath: str) -> str:
    """
    Concatenate all fasta files into a temporary fasta file.

    :param dirpath: directory containing fasta files
    :type dirpath: str
    :return: the temporary filename
    :rtype: str
    """
    fasta_files = glob(join(dirpath, "*.fsa"))
    fasta_files += glob(join(dirpath, "*.fasta"))
    records = []
    for fsa in fasta_files:
        records += list(SeqIO.parse(fsa, "fasta"))
    return records_to_tmpfile(records)


def _grouped_by_id(records):
    """
    Return a dictionary of records grouped by their record_id
    :param records:
    :type records:
    :return:
    :rtype:
    """
    rec_id_dict = {}
    for r in records:
        rec_id_dict.setdefault(r.id, list()).append(r)


def force_unique_record_ids(records, use_uuid=False):
    """
    Force unique ids for all records in the list

    :param records: list of SeqRecords
    :type records: list
    :param use_uuid: If True, will append a unique identifier to the front of the record id. Else, enumerate the
    records.
    :type use_uuid: bool
    :return: the records
    :rtype: list
    """
    rec_id_dict = {}
    for r in records:
        rec_id_dict.setdefault(r.id, list()).append(r)
    for k, v in rec_id_dict.items():
        if len(v) > 1:
            for i, r in enumerate(v):
                if use_uuid:
                    i = str(uuid4())
                r.id = "({})__{}".format(i, r.id)
    return records


def load_glob(
    path: str, format_str: str, recursive=False, force_unique_ids=False
) -> List[SeqRecord]:
    """
    Load SeqRecords from a glob-like path.

    :param path: glob-like filepath
    :type path: str
    :param format_str: Fileformat. e.g. 'genbank', 'fasta'
    :type format_str: basestring
    :param recursive: whether to search recusively in the glob path.
    :type recursive: bool
    :param force_unique_ids: If True, force records to have unique ids
    :type force_unique_ids: bool
    :return: the list of SeqRecords
    :rtype: list
    """
    records = []
    filenames = glob(path, recursive=recursive)
    for f in filenames:
        new_records = list(SeqIO.parse(f, format=format_str))
        for rec in new_records:
            rec.from_file = f
            records.append(rec)

    if force_unique_ids:
        force_unique_record_ids(records)
    return records


def load_fasta_glob(path, recursive=False, force_unique_ids=False):
    """
    Load SeqRecords from a glob-like path containing fasta files.

    :param path: glob-like filepath
    :type path: str
    :param recursive: whether to search recusively in the glob path.
    :type recursive: bool
    :param force_unique_ids: If True, force records to have unique ids
    :type force_unique_ids: bool
    :return: the list of SeqRecords
    :rtype: list
    """
    return load_glob(
        path, "fasta", recursive=recursive, force_unique_ids=force_unique_ids
    )


def load_genbank_glob(path, recursive=False, force_unique_ids=False):
    """
    Load SeqRecords from a glob-like path containing genbank files.

    :param path: glob-like filepath
    :type path: str
    :param recursive: whether to search recusively in the glob path.
    :type recursive: bool
    :param force_unique_ids: If True, force records to have unique ids
    :type force_unique_ids: bool
    :return: the list of SeqRecords
    :rtype: list
    """
    records = load_glob(
        path, "genbank", recursive=recursive, force_unique_ids=force_unique_ids
    )
    for r in records:
        if C.TOPOLOGY in r.annotations and r.annotations[C.TOPOLOGY] == C.CIRCULAR:
            r.annotations[C.CIRCULAR] = True
        else:
            r.annotations[C.CIRCULAR] = False
    return records


def clean_records(records: List[SeqRecord]) -> None:
    """
    Remove features for records that have no location.
    """
    for r in records:
        r.features = [f for f in r.features if f.location is not None]


def is_circular(record: SeqRecord) -> bool:
    """
    Returns whether a "topology" annotation was found in the SeqRecord and
    whether this topolgy indicates that the sequence is circular.

    :param record: the SeqRecord
    :type record: SeqRecord
    :return: whether the SeqRecord is circular
    :rtype: bool
    """
    annotations = {k.lower(): v for k, v in record.annotations.items()}
    return (
        annotations.get(C.CIRCULAR.lower(), False) is True
        or annotations.get(C.LINEAR.lower(), True) is False
        or annotations.get(C.TOPOLOGY.lower(), C.LINEAR).lower() == C.CIRCULAR.lower()
    )


def make_linear(records: List[SeqRecord]) -> List[SeqRecord]:
    """
    Annotates the SeqRecords as linear by adding the 'topology' annotation and
    setting it to 'linear'

    :param records:
    :type records:
    :return: the records
    :rtype: list
    """
    for r in records:
        if C.CIRCULAR in r.annotations:
            r.annotations[C.CIRCULAR] = False
        elif C.LINEAR in r.annotations:
            r.annotations[C.LINEAR] = True
        else:
            r.annotations[C.TOPOLOGY] = C.LINEAR
    return records


def make_circular(records: List[SeqRecord]) -> List[SeqRecord]:
    """
    Annotates the SeqRecords as linear by adding the 'topology' annotation and
    setting it to 'linear'

    :param records:
    :type records:
    :return: the records
    :rtype: list
    """
    for r in records:
        if C.CIRCULAR in r.annotations:
            r.annotations[C.CIRCULAR] = True
        elif C.LINEAR in r.annotations:
            r.annotations[C.LINEAR] = False
        else:
            r.annotations[C.TOPOLOGY] = C.CIRCULAR
    return records
